#pragma once

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctime>
#include <limits.h>

#define TEST

#ifdef TEST
#define OFFLINE//Tests without Slide web service
//#define PRINT_POSITIONS
#endif
#define COMPRESS_BOARD//board==short[5] instead of int[5][5] or char[5][5]
//#define EQUALIZE_COLUMNS//This strategy at quiet positions improves magically game score
//#define CUTOFF//Speculative cutoff: gives lower scores, without essentially decreasing search time
#define ALPHA_BETA

using namespace System;

const int BOARD_SIZE=5,FIG_BITS=3,MAX_MOVES=200,MAX_DEPTH=16,MIN_DUMMY=5;

typedef unsigned short ROW;

//Now in Insert.cpp:
//ROW Mask[]={7,7<<3,7<<6,7<<9,7<<12};
//ROW Mask1[]={~7,~(7<<3),~(7<<6),~(7<<9),~(7<<12)};
extern ROW Mask[],Mask1[];

enum DIR{HOR,VER};
//enum OCTON{_0,_1,_2,_3,_4,_5,_6,_7};//Enumeration of "pictures"

struct segment{
	int row,col;//First point
	int len;
	DIR dir;//horizontal or vertical
	segment():row(0),col(0),len(0),dir(HOR){}
	segment&operator=(segment seg){
		row=seg.row,col=seg.col,len=seg.len,dir=seg.dir;
		return *this;
	}
};

//Position row compressed to 16 bits (3 bits are enough to code one picture number)
//In fact search with compressed rows is 1%-2% slower than that with plain arrays.
//But in code compiled with arrays there are some strange bugs while searching at depth 6
//if the program runs from console (everything is OK if it starts from IDE under debugger (F5)).
struct Row{
#ifdef COMPRESS_BOARD
	ROW row;
	
	Row():row(0){}

	void set(int col,int n){row&=Mask1[col];row|=((ROW)n&7)<<FIG_BITS*col;}
	
	int get(int col)const{return 7&(row>>(FIG_BITS*col));}

	Row&operator<<=(int shift){row<<=FIG_BITS*shift;return *this;}

	Row&operator>>=(int shift){row>>=FIG_BITS*shift;return *this;}

	//proxy for an element; the trick is borrowed from "bitset", to define good operator[].
	struct reference{
		friend struct Row;
		reference&operator=(int val){//assign int to element
			PRow->set(ind,val);
			return *this;
		}

		reference&operator=(const reference&Ref){//assign reference to element
			PRow->set(ind,Ref.PRow->get(Ref.ind));
			return *this;
		}

		operator int()const{return PRow->get(ind);}//Without this int i=row[j] is bad!

	private:
		Row*PRow;
		int ind;
		reference(Row&R,int i):PRow(&R),ind(i){}//construct from Row reference and index
	};
		
	int operator[](int col)const{return get(col);}

	reference operator[](int col){return reference(*this,col);}

#if 0
	bool operator==(const Row&r)const{
		//return r.row==row;//Bad! 16th bit is irrelevant!
		return (r.row&0x7FFF)==(row&0x7FFF);//This is bad as well! Why?
	}

	bool operator!=(const Row&r)const{
		return !(r.row==row);
	}
#endif

#else
	
	int row[BOARD_SIZE];

	Row(){for(int i=0;i<BOARD_SIZE;i++)row[i]=0;}

	void set(int col,int n){row[col]=n;}
	
	int get(int col)const{return row[col];}

	int operator[](int col)const{return row[col];}

	int& operator[](int col){return row[col];}

	Row&operator<<=(int shift){
		for(int i=BOARD_SIZE-1;i>=shift;i--)row[i]=row[i-shift];
		return *this;
	}

	Row&operator>>=(int shift){
		for(int i=0;i<BOARD_SIZE-shift;i++)row[i]=row[i+shift];
		return *this;
	}

#endif

	bool operator==(const Row&r)const{
		for(int c=0;c<BOARD_SIZE;c++){
			//if(r.row[c]!=row[c])return false;//Does not work!
			//if((int)r.get(c)!=(int)get(c))return false;//This is bad as well!
			int val1=r.get(c),val2=get(c);
			//int val1=r[c],val2=(*this)[c];
			if(val1!=val2)return false;//This works. Sometimes.
		}
		return true;
	}

	bool operator!=(const Row&r)const{
		return !(r==*this);
	}

	
	Row(array<int>^board){
		for(int i=0;i<BOARD_SIZE;i++)(*this)[i]=board[i];
	}

	Row&operator=(array<int>^board){
		for(int i=0;i<BOARD_SIZE;i++)(*this)[i]=board[i];
		return *this;
	}

	void Print(){
		for(int c=0;c<BOARD_SIZE;c++)
			printf("%d",(int)(*this)[c]);
		//printf("\n");
	}

};

struct EVAL{
	int score;
	int deleted;//number of deleted figures: the less the better
	EVAL(int score=0,int deleted=0):score(score),deleted(deleted){}
	EVAL&operator=(int sc){score=sc,deleted=0;return*this;}
	EVAL&operator=(const EVAL E){score=E.score,deleted=E.deleted;return*this;}
	EVAL&operator+=(const EVAL E){score+=E.score,deleted+=E.deleted;return*this;}
	EVAL&operator*(int m){score=m*score;return*this;}
	EVAL&operator*=(int m){score*=m;return*this;}
	//This is to be used to maximize my score
	bool operator>(const EVAL&E)const{
		if(score!=E.score)return score>E.score;
		return deleted<E.deleted;
	}
#if 1
	//This is to be used to minimize opponent's score!
	bool operator<(const EVAL&E)const{//< != !>= !!!
		if(score!=E.score)return score<E.score;
		return deleted<E.deleted;
	}
#endif
	bool operator==(const EVAL&E)const{
		return score==E.score&&deleted==E.deleted;
	}
	bool operator>=(const EVAL&E)const{
		return *this>E||*this==E;
	}
	bool operator<=(const EVAL&E)const{
		return *this<E||*this==E;
	}
	operator int()const{return score;}
};

struct EVALOK{
	int score[2];
	int deleted;//number of deleted figures: the less the better
	EVALOK(int Score=0,int ScoreOp=0,int deleted=0):deleted(deleted){score[0]=Score,score[1]=ScoreOp;}
	EVALOK&operator=(int sc){score[0]=score[1]=deleted=sc;return*this;}//for use with sc==0 only
	EVALOK&operator=(int *sc){score[0]=sc[0],score[1]=sc[1],deleted=0;return*this;}
	EVALOK&operator=(const EVALOK E){score[0]=E.score[0],score[1]=E.score[1],deleted=E.deleted;return*this;}
	EVALOK&operator+=(const EVALOK E){score[0]+=E.score[0],score[1]+=E.score[1],deleted+=E.deleted;return*this;}
	EVALOK&operator-=(const EVALOK E){score[0]-=E.score[0],score[1]-=E.score[1],deleted-=E.deleted;return*this;}
	bool operator>(const EVALOK&E)const{
		if(score[0]!=E.score[0])return score[0]>E.score[0];
		if(deleted!=E.deleted)return deleted<E.deleted;
		return score[1]<E.score[1];
	}
	bool operator<(const EVALOK&E)const{//Here < == !>=
		if(score[0]!=E.score[0])return score[0]<E.score[0];
		if(deleted!=E.deleted)return deleted>E.deleted;
		return score[1]>E.score[1];
	}
	bool operator==(const EVALOK&E)const{
		return score[0]==E.score[0]&&score[1]==E.score[1]&&deleted==E.deleted;
	}
	bool operator>=(const EVALOK&E)const{
		return *this>E||*this==E;
	}
	bool operator<=(const EVALOK&E)const{
		return *this<E||*this==E;
	}
	//operator int()const{return score[0];}
};

extern EVALOK INFPLUS,INFMINUS;

//Search "engines":
#define DUMMY		0
#define RANDOM	1 
#define EVAL2		2//2-tier evaluation: Max score -- Min Opponent score (Not score - opScore!!!)
#define EVAL2X	3//Modded 2-tier
#define EVAL3		4//3-tier evaluation: Max score -- Min deleted pics -- Min Opponent score
#define EVAL3X	5//Modded 3-tier
#define EVAL2XX	6//Experiments with cutoffs, forced play, etc.
#define EVAL3XX	7//Experiments with cutoffs, forced play, etc.
#define EVAL4X	8//Eval3X with more exact opponent evaluation (deleted -- score instead of score -- deleted)
#define EVAL4XX	9//Eval3XX with fixed bugs and partial alpha-beta
#define EVAL5		10//True 3-tier evaluation with alpha-beta-good results but theoretically incorrect
#define EVAL5X	11//True 3-tier evaluation with alpha-beta

extern char*EngineNames[16];

struct Engine{
	int strat,depth;//search strategy and search depth
	Engine(int st=0,int dp=0):strat(st),depth(st==DUMMY||st==RANDOM?0:dp){}
};

struct Board{
	Row pos[BOARD_SIZE];
	Row Height;//In fact it is the complement to the heights of columns

	Board(){
		//for(int r=0;r<BOARD_SIZE;r++)pos[r]=0;
		//Height=0;
	}
	
	//Generate random board
	Board(int P,int seed=0){
		if(seed!=0)srand(seed);
		segment seg;
		do{
			for(int r=0;r<BOARD_SIZE;r++)
			for(int c=0;c<BOARD_SIZE;c++)pos[r][c]=rand()%P+1;
#ifdef COMPRESS_BOARD
			Height.row=0;
#else
			for(int c=0;c<BOARD_SIZE;c++)Height.row[c]=0;
#endif
			Segment2Remove(seg);
#ifdef TEST1
		printf("Random Initial position:\n\n");
		Print();
		printf("seg.len=%d\n\n",seg.len);
#endif
		}while(seg.len>0);
#ifdef TEST
		printf("Random Initial position:\n\n");
		Print();
#endif
	}

	Board&operator=(const Board&board){
		for(int r=0;r<BOARD_SIZE;r++)pos[r]=board.pos[r];
		Height=board.Height;
		return *this;
	}

	Board(array<array<int>^>^board){
		int r,c;
		for(r=0;r<BOARD_SIZE;r++)pos[r]=board[r];
		for(c=0;c<BOARD_SIZE;c++){
			r=0;
			while(r<BOARD_SIZE&&board[r][c]==0)r++;
			Height[c]=r;
		}
	}

	Board&operator=(array<array<int>^>^board){
		int r,c;
		for(r=0;r<BOARD_SIZE;r++)pos[r]=board[r];
		for(c=0;c<BOARD_SIZE;c++){
			r=0;
			while(r<BOARD_SIZE&&board[r][c]==0)r++;
			Height[c]=r;
		}
		return *this;
	}

//#define Height(col) Height.get(col)
#define Height(col) Height[col]
//#define setHeight(col,val) Height.set(col,val)
#define setHeight(col,val) Height[col]=val
#define IsEmpty(r,c) (r<Height[c])
	
	void set(int row,int col,int n){pos[row][col]=n;}
	
	int get(int row,int col)const{return pos[row][col];}
	
	int kickR(int r,int val){
		int c,c1;
		for(c=0;c<BOARD_SIZE;c++)if(IsEmpty(r,c))break;
		if(c==BOARD_SIZE){pos[r]<<=1,pos[r][0]=val;return 1;}
		int h=Height[c]-1;
		Height[c]=h;
		if(c==0)pos[h][c]=val;
		else{
			pos[h][c]=pos[r][c-1];
			for(c1=c-1;c1>0;c1--)pos[r][c1]=pos[r][c1-1];
			pos[r][0]=val;
		}
		return 0;
	}
	
	EVAL kickL(int r,int val){
		int c,c1;
		for(c=BOARD_SIZE-1;c>=0;c--)if(IsEmpty(r,c))break;
		if(c==-1){pos[r]>>=1,pos[r][BOARD_SIZE-1]=val;return 1;}
		int h=Height[c]-1;
		Height[c]=h;
		if(c==BOARD_SIZE-1)pos[h][c]=val;
		else{
			pos[h][c]=pos[r][c+1];
			for(c1=c+1;c1<BOARD_SIZE-1;c1++)pos[r][c1]=pos[r][c1+1];
			pos[r][BOARD_SIZE-1]=val;
		}
		return 0;
	}
	
	EVAL kickD(int c,int val){
		int h=Height[c]-1;
		if(h>=0){
			pos[h][c]=val,Height[c]=h;
			return 0;
		}
		else{
			for(int r=BOARD_SIZE-1;r>0;r--)pos[r][c]=pos[r-1][c];
			pos[0][c]=val;
			return 1;
		}
	}

	void Segment2Remove(segment&seg)const{
		seg.row=seg.col=seg.len=0;seg.dir=HOR;
		for(int r=0;r<BOARD_SIZE;r++){
			for(int c=0,len=0;c<3;){
				if(IsEmpty(r,c)){c++;continue;}
				if(IsEmpty(r,c+1)){c+=2;continue;}
				if(pos[r][c]!=pos[r][c+1]){c++;continue;}
				if(IsEmpty(r,c+2))break;
				if(pos[r][c+1]!=pos[r][c+2]){c+=2;continue;}
				c+=2,len=3;
				if(c<4&&!IsEmpty(r,c+1)&&pos[r][c]==pos[r][c+1])c++,len++;
				if(c<4&&!IsEmpty(r,c+1)&&pos[r][c]==pos[r][c+1])c++,len++;
				if(len>seg.len)
					seg.row=r,seg.col=c-len+1,seg.len=len,seg.dir=HOR;
				//This check slows down a bit:
				//if(seg.len==5)return;//The champ is found
				break;//There may be only one segment in a row!
			}
		}
		if(seg.len==BOARD_SIZE)return;
		//Look for longer vertical segments
		for(int c=0;c<BOARD_SIZE;c++){
			int r=Height[c];
			if(r>2)continue;//Speeds up a bit
			for(int len=0;r<3;){//IsEmty() check is superfluous for vertical segments
				if(pos[r][c]!=pos[r+1][c]){r++;continue;}
				if(pos[r+1][c]!=pos[r+2][c]){
					r+=2;
					if(seg.len>2)break;//No chance to find seg with seg.len>3 in this column
					continue;
				}
				r+=2,len=3;
				if(r<4&&pos[r][c]==pos[r+1][c])r++,len++;
				if(r<4&&pos[r][c]==pos[r+1][c])r++,len++;
				if(len>seg.len||len==seg.len&&seg.dir==VER&&seg.row>r-len+1)
					seg.row=r-len+1,seg.col=c,seg.len=len,seg.dir=VER;
				//This check slows down a bit:
				//if(seg.len==5)return;//The champ is found
				break;//There may be only one segment in a column!
			}
		}
	}

	void Remove(const segment&seg){
		if(seg.dir==HOR){
			for(int c=seg.col;c<seg.col+seg.len;c++){
				for(int r=seg.row-1;!(IsEmpty(r,c));r--)pos[r+1][c]=pos[r][c];
				int h=Height[c]+1;
				Height[c]=h;
			}
		}
		else{
			const int&c=seg.col;
			for(int r=seg.row-1;!(IsEmpty(r,c));r--)pos[r+seg.len][c]=pos[r][c];
			int h=Height[c]+seg.len;
			Height[c]=h;
		}
	}
	
	EVAL RemoveSegment(){
		EVAL eval;
		segment seg;
		Segment2Remove(seg);
		if(seg.len==0)return eval;
		Remove(seg);
		eval.score=10*(seg.len-2);
		eval.deleted=seg.len;
		return eval;
	}

	EVAL RemoveAll(){
		EVAL eval=0,deltaeval;
		int m=0;
		do{
#if 1
			deltaeval=RemoveSegment();
			deltaeval*=++m;
#else
			deltaeval=++m*RemoveSegment();
#endif
			eval+=deltaeval;
#ifdef TEST1
			if(deltaeval.score>0){
				Print();
				printf("Heights: ");
				Height.Print();printf("\n");
			}
#endif
		}while(deltaeval.score>0);
		return eval;
	}

	//Moves are numbered internally from 0 to 14
	//So every sequence of depth moves can be represented by depth-digit 15-ary number
	//(if there is no wildcards at the end of the sequence)
	int MakeMove(int move,int val){
		if(move<5)
			kickR(4-move,val);
		else
		if(move<10)
			kickD(move-5,val);
		else
			kickL(move-10,val);
		return RemoveAll();
	}

	EVAL MakeMoveX(int move,int val){
		EVAL eval=0;
		int deleted;
		if(move<5)
			deleted=kickR(4-move,val);
		else
		if(move<10)
			deleted=kickD(move-5,val);
		else
			deleted=kickL(move-10,val);
		eval.deleted=deleted;
		eval+=RemoveAll();
		return eval;
	}

	EVALOK MakeMoveX(int move,int val,int player){
		EVAL eval=MakeMoveX(move,val);
		EVALOK evalOK=0;
		evalOK.deleted=eval.deleted,evalOK.score[player&1]=eval.score;
		return evalOK;
	}

	void Print(){
		int val;
		for(int r=0;r<BOARD_SIZE;r++){
			for(int c=0;c<BOARD_SIZE;c++)
				if(!(IsEmpty(r,c)))printf("%d",val=pos[r][c]);//Strange bug if one writes simply pos[r][c]
				else printf("*");
			printf("\n");
		}
		printf("\n");
	}

	void EvalDummy(int&bestmove,int dummy=5){bestmove=dummy;}

	void EvalRand(int&bestmove){bestmove=rand()%15;}
	
	int Eval(int&bestmove,int pict,bool extraCode=false){
		int eval=0;
		bestmove=0;
		for(int m=0;m<15;m++){
			if(4-m>=0&&IsEmpty(4-m,0))continue;//this move is duplicated by move 5
			if(m-10>=0&&IsEmpty(m-10,BOARD_SIZE-1))continue;//this move is duplicated by move 9
			Board Board1=*this;
			int eval1=Board1.MakeMove(m,pict);
			if(eval1>eval)eval=eval1,bestmove=m;
		}
		if(extraCode&&eval==0){//Make "vertical" move into column with minimal height, increasing the number of non-empty squares
			int h=0,c0=0;
			for(int c=0;c<BOARD_SIZE;c++)
				if(Height[c]>h)h=Height[c],c0=c;
			bestmove=5+c0;
		}
		return eval;
	}

	void Eval(int&bestmove,int&eval,int&evalOp,int depth,array<int>^squares,int k,int dummyOp=-1,bool extraCode=false){
		if(depth<1||k>=squares->Length){eval=evalOp=bestmove=0;return;}
		if(depth==1){eval=Eval(bestmove,squares[k],extraCode);evalOp=0;return;}
		eval=bestmove=evalOp=0;
		for(int m=0;m<15;m++){
			if(4-m>=0&&IsEmpty(4-m,0))continue;//this move is duplicated by move 5
			if(m-10>=0&&IsEmpty(m-10,BOARD_SIZE-1))continue;//this move is duplicated by move 9
			Board Board1=*this;
			int eval1=Board1.MakeMove(m,squares[k]),eval2min=1000,evalOp=0,bestevalOp=1000;
			if(k+1>=squares->Length)goto COMPARE;
			if(dummyOp>-1){
				Board1.MakeMove(dummyOp,squares[k+1]);
				int bestmove1,eval2;
				Board1.Eval(bestmove1,eval2,evalOp,depth-1,squares,k+2,dummyOp);
				eval1+=eval2;
				goto COMPARE;
			}
			for(int m1=0;m1<15;m1++){
				if(4-m1>=0&&IsEmpty(4-m1,0))continue;//this move is duplicated by move 5
				if(m1-10>=0&&IsEmpty(m1-10,BOARD_SIZE-1))continue;//this move is duplicated by move 9
				Board Board2=Board1;
				int evalOp1=Board2.MakeMove(m1,squares[k+1]);
				if(evalOp1>evalOp)evalOp=evalOp1;//,bestmoveOp=m1;
				int bestmove1;
				int eval2;
				Board2.Eval(bestmove1,eval2,evalOp,depth-2,squares,k+2,dummyOp);
				if(eval2min>eval2)eval2min=eval2;
			}
			eval1+=eval2min;
COMPARE:
			if(eval1>eval)eval=eval1,bestmove=m,bestevalOp=evalOp;
			else
			if(eval1==eval&&evalOp<bestevalOp)eval=eval1,bestmove=m,bestevalOp=evalOp;
		}
		if(extraCode&&eval==0){//Make "vertical" move into column with minimal height, increasing the number of non-empty squares
			int h=0,c0=0;
			for(int c=0;c<BOARD_SIZE;c++)
				if(Height[c]>h)h=Height[c],c0=c;
			bestmove=5+c0;
		}
		//return eval;
	}

	EVAL EvalX(int&bestmove,int pict,bool extraCode=false){
		EVAL eval=0;
		bestmove=0;
		for(int m=0;m<15;m++){
			if(4-m>=0&&IsEmpty(4-m,0))continue;//this move is duplicated by move 5
			if(m-10>=0&&IsEmpty(m-10,BOARD_SIZE-1))continue;//this move is duplicated by move 9
			Board Board1=*this;
			EVAL eval1=Board1.MakeMoveX(m,pict);
			if(eval1>eval)eval=eval1,bestmove=m;
		}
		if(extraCode&&eval.score==0){//Make "vertical" move into column with minimal height, increasing the number of non-empty squares
			int h=0,c0=0;
			for(int c=0;c<BOARD_SIZE;c++)
				if(Height[c]>h)h=Height[c],c0=c;
			bestmove=5+c0;
		}
		return eval;
	}

	void EvalX(int&bestmove,EVAL&eval,EVAL&evalOp,int depth,array<int>^squares,int k,int dummyOp=-1,bool extraCode=false)
	{
		if(depth<1||k>=squares->Length){eval=evalOp=bestmove=0;return;}
		if(depth==1){eval=EvalX(bestmove,squares[k],extraCode);evalOp=0;return;}
		eval=evalOp=0;bestmove=0;
		for(int m=0;m<15;m++){
			if(4-m>=0&&IsEmpty(4-m,0))continue;//this move is duplicated by move 5
			if(m-10>=0&&IsEmpty(m-10,BOARD_SIZE-1))continue;//this move is duplicated by move 9
			Board Board1=*this;
			EVAL eval1=Board1.MakeMoveX(m,squares[k]),eval2min=1000,evalOp=0,bestevalOp=1000;
			if(k+1>=squares->Length)goto COMPARE;
			if(dummyOp>-1){
				Board1.MakeMoveX(dummyOp,squares[k+1]);
				int bestmove1;
				EVAL eval2;
				Board1.EvalX(bestmove1,eval2,evalOp,depth-1,squares,k+2,dummyOp);//Extra code here is not necessary
				eval1+=eval2;
				goto COMPARE;
			}
			for(int m1=0;m1<15;m1++){
				if(4-m1>=0&&IsEmpty(4-m1,0))continue;//this move is duplicated by move 5
				if(m1-10>=0&&IsEmpty(m1-10,BOARD_SIZE-1))continue;//this move is duplicated by move 9
				Board Board2=Board1;
				EVAL evalOp1=Board2.MakeMoveX(m1,squares[k+1]);
				if(evalOp1>evalOp)evalOp=evalOp1;
				int bestmove1;
				EVAL eval2;
				Board2.EvalX(bestmove1,eval2,evalOp,depth-2,squares,k+2,dummyOp);
				if(eval2min>eval2)eval2min=eval2;
			}
			eval1+=eval2min;
COMPARE:
			if(eval1>eval)
				eval=eval1,bestmove=m,bestevalOp=evalOp;
			else
			if(eval1==eval&&evalOp<bestevalOp)
				bestmove=m,bestevalOp=evalOp;
		}
		if(eval.score==0&&extraCode){//Make "vertical" move into column with minimal height, increasing the number of non-empty squares
			int h=0,c0=0;
			for(int c=0;c<BOARD_SIZE;c++)
				if(Height[c]>h)h=Height[c],c0=c;
			bestmove=5+c0;
		}
	}

	void Eval2XX(int&bestmove,int&eval,int&evalOp,int depth,array<int>^squares,int k,int dummyOp=-1,bool extraCode=false){
		if(depth<1||k>=squares->Length){eval=evalOp=bestmove=0;return;}
		if(depth==1){eval=Eval(bestmove,squares[k],extraCode);evalOp=0;return;}
		eval=-2000,bestmove=evalOp=0;
		for(int m=0;m<15;m++){
			if(4-m>=0&&IsEmpty(4-m,0))continue;//this move is duplicated by move 5
			if(m-10>=0&&IsEmpty(m-10,BOARD_SIZE-1))continue;//this move is duplicated by move 9
			Board Board1=*this;
			int eval1=Board1.MakeMove(m,squares[k]),eval2min=2000,bestevalOp=0;
			if(k+1>=squares->Length)goto COMPARE;
			if(dummyOp>-1){
				Board1.MakeMove(dummyOp,squares[k+1]);
				int bestmove1,eval2;
				Board1.Eval(bestmove1,eval2,evalOp,depth-1,squares,k+2,dummyOp);
				eval1+=eval2;
				goto COMPARE;
			}
			for(int m1=0;m1<15;m1++){
				if(4-m1>=0&&IsEmpty(4-m1,0))continue;//this move is duplicated by move 5
				if(m1-10>=0&&IsEmpty(m1-10,BOARD_SIZE-1))continue;//this move is duplicated by move 9
				Board Board2=Board1;
				int evalOp1=Board2.MakeMove(m1,squares[k+1]),evalOp2;
#ifdef CUTOFF
				//Cutoff at even depth?!
				if(evalOp1<evalOp/2)continue;
#endif
				int bestmove1;
				int eval2;
				Board2.Eval(bestmove1,eval2,evalOp2,depth-2,squares,k+2,dummyOp);
				if(evalOp1+evalOp2>evalOp)bestevalOp=evalOp1+evalOp2;
				if(eval2min>eval2)eval2min=eval2;
			}
			eval1+=eval2min;
COMPARE:
			if(eval1>eval)eval=eval1,bestmove=m,evalOp=bestevalOp;
			else
			if(eval1==eval&&evalOp>bestevalOp)eval=eval1,bestmove=m,evalOp=bestevalOp;
		}
#ifdef EQUALIZE_COLUMNS
		if(extraCode&&eval==0){//Make "vertical" move into column with minimal height, increasing the number of non-empty squares
			int h=0,c0=0;
			for(int c=0;c<BOARD_SIZE;c++)
				if(Height[c]>h)h=Height[c],c0=c;
			bestmove=5+c0;
		}
#endif
	}

	//Contest version
	void Eval3XX(int&bestmove,EVAL&eval,EVAL&evalOp,int depth,array<int>^squares,int k,int dummyOp=-1,bool extraCode=false)
	{
		if(depth<1||k>=squares->Length){eval=evalOp=bestmove=0;return;}
		if(depth==1){eval=EvalX(bestmove,squares[k],extraCode);evalOp=0;return;}
		eval=-2000,evalOp=0;bestmove=0;
		for(int m=0;m<15;m++){
			if(4-m>=0&&IsEmpty(4-m,0))continue;//this move is duplicated by move 5
			if(m-10>=0&&IsEmpty(m-10,BOARD_SIZE-1))continue;//this move is duplicated by move 9
			Board Board1=*this;
			EVAL eval1=Board1.MakeMoveX(m,squares[k]),eval2min=2000,bestevalOp=0;
#ifdef CUTOFF
				//Cutoff at odd depth?!
				if(eval1.score<evalOp.score/4)continue;
#endif
			if(k+1>=squares->Length)goto COMPARE;
			if(dummyOp>-1){
				Board1.MakeMoveX(dummyOp,squares[k+1]);
				int bestmove1;
				EVAL eval2;
				//Board1.EvalX(bestmove1,eval2,evalOp,depth-1,squares,k+2,dummyOp);//Extra code here is not necessary
				Board1.Eval3XX(bestmove1,eval2,evalOp,depth-1,squares,k+2,dummyOp);//Extra code here is not necessary
				eval1+=eval2;
				goto COMPARE;
			}
			for(int m1=0;m1<15;m1++){
				if(4-m1>=0&&IsEmpty(4-m1,0))continue;//this move is duplicated by move 5
				if(m1-10>=0&&IsEmpty(m1-10,BOARD_SIZE-1))continue;//this move is duplicated by move 9
				Board Board2=Board1;
				EVAL evalOp1=Board2.MakeMoveX(m1,squares[k+1]),evalOp2;
#ifdef CUTOFF
				//Cutoff at even depth?!
				if(evalOp1.score<bestevalOp.score/4)continue;
#endif
				int bestmove1;
				EVAL eval2;
				Board2.EvalX(bestmove1,eval2,evalOp2,depth-2,squares,k+2,dummyOp);
				if(evalOp1+evalOp2>bestevalOp)bestevalOp=evalOp1+evalOp2;
				if(eval2min>eval2)eval2min=eval2;
			}
			eval1+=eval2min;
COMPARE:
			if(eval1>eval)
				eval=eval1,bestmove=m,evalOp=bestevalOp;
			else
			if(eval1==eval&&evalOp>bestevalOp)
				bestmove=m,evalOp=bestevalOp;
		}
#ifdef EQUALIZE_COLUMNS
		if(eval.score==0&&extraCode){//Make "vertical" move into column with minimal height, increasing the number of non-empty squares
			int h=0,c0=0;
			for(int c=0;c<BOARD_SIZE;c++)
				if(Height[c]>h)h=Height[c],c0=c;
			bestmove=5+c0;
		}
#endif
	}

	//Best of buggy versions.
	void Eval4X(int&bestmove,EVAL&eval,EVAL&evalOp,int depth,array<int>^squares,int k,int dummyOp=-1,bool extraCode=false)
	{
		if(depth<1||k>=squares->Length){eval=evalOp=bestmove=0;return;}
		if(depth==1){eval=EvalX(bestmove,squares[k],extraCode);evalOp=0;return;}
		eval=-2000,evalOp=0;bestmove=0;
		for(int m=0;m<15;m++){
			if(4-m>=0&&IsEmpty(4-m,0))continue;//this move is duplicated by move 5
			if(m-10>=0&&IsEmpty(m-10,BOARD_SIZE-1))continue;//this move is duplicated by move 9
			Board Board1=*this;
			EVAL eval1=Board1.MakeMoveX(m,squares[k]),eval2min=2000,bestevalOp=0;
			if(k+1>=squares->Length)goto COMPARE;
			if(dummyOp>-1){
				Board1.MakeMoveX(dummyOp,squares[k+1]);
				int bestmove1;
				EVAL eval2;
				Board1.Eval4X(bestmove1,eval2,evalOp,depth-1,squares,k+2,dummyOp);//Extra code here is not necessary
				eval1+=eval2;
				goto COMPARE;
			}
			for(int m1=0;m1<15;m1++){
				if(4-m1>=0&&IsEmpty(4-m1,0))continue;//this move is duplicated by move 5
				if(m1-10>=0&&IsEmpty(m1-10,BOARD_SIZE-1))continue;//this move is duplicated by move 9
				Board Board2=Board1;
				EVAL evalOp1=Board2.MakeMoveX(m1,squares[k+1]),evalOp2;
				int bestmove1;
				EVAL eval2;
				Board2.EvalX(bestmove1,eval2,evalOp2,depth-2,squares,k+2,dummyOp);
				//Board2.Eval4X(bestmove1,eval2,evalOp2,depth-2,squares,k+2,dummyOp);//This is worse. Why?
				if(evalOp1+evalOp2>bestevalOp)bestevalOp=evalOp1+evalOp2;
				if(eval2min>eval2)eval2min=eval2;
#ifdef ALPHA_BETA
				//Partial alpha-beta: though incorrect, beats Eval3XX() and runs 5 times faster!
				EVAL curr=eval1+eval2;
				if(eval>curr||eval==curr)break;
				//if(eval==curr)break;
#endif
			}
			eval1+=eval2min;
COMPARE:
			if(eval1>eval)
				eval=eval1,bestmove=m,evalOp=bestevalOp;
			else
				if(eval1==eval&&bestevalOp<evalOp)
				bestmove=m,evalOp=bestevalOp;
		}
	}

	//Correct minimax with (partial yet) alpha-beta. Speed up 4 times
	void Eval4XX(int&bestmove,EVAL&eval,EVAL&evalOp,int depth,array<int>^squares,int k,int dummyOp=-1,bool extraCode=false)
	{
		if(depth<1||k>=squares->Length){eval=evalOp=bestmove=0;return;}
		if(depth==1){eval=EvalX(bestmove,squares[k],extraCode);evalOp=0;return;}
		eval=-2000,evalOp=0;bestmove=0;
		for(int m=0;m<15;m++){
			if(4-m>=0&&IsEmpty(4-m,0))continue;//this move is duplicated by move 5
			if(m-10>=0&&IsEmpty(m-10,BOARD_SIZE-1))continue;//this move is duplicated by move 9
			Board Board1=*this;
			EVAL eval1=Board1.MakeMoveX(m,squares[k]),eval2min=2000,bestevalOp=0;
			if(k+1>=squares->Length)goto COMPARE;
			if(dummyOp>-1){
				Board1.MakeMoveX(dummyOp,squares[k+1]);
				int bestmove1;
				EVAL eval2;
				Board1.Eval4XX(bestmove1,eval2,evalOp,depth-1,squares,k+2,dummyOp);
				eval1+=eval2;
				goto COMPARE;
			}
			for(int m1=0;m1<15;m1++){
				if(4-m1>=0&&IsEmpty(4-m1,0))continue;//this move is duplicated by move 5
				if(m1-10>=0&&IsEmpty(m1-10,BOARD_SIZE-1))continue;//this move is duplicated by move 9
				Board Board2=Board1;
				EVAL evalOp1=Board2.MakeMoveX(m1,squares[k+1]),evalOp2;
				int bestmove1;
				EVAL eval2;
				Board2.Eval4XX(bestmove1,eval2,evalOp2,depth-2,squares,k+2,dummyOp);
				if(eval2min>eval2)eval2min=eval2,bestevalOp=evalOp1+evalOp2;
				else
				if(eval2min==eval2&&evalOp1+evalOp2>bestevalOp)bestevalOp=evalOp1+evalOp2;
#ifdef ALPHA_BETA
				//Partial alpha-beta:
				EVAL curr=eval1+eval2,currOp=evalOp1+evalOp2;
				if(eval>curr||eval==curr&&currOp>=evalOp)break;
#endif
			}
			eval1+=eval2min;
COMPARE:
			if(eval1>eval)
				eval=eval1,bestmove=m,evalOp=bestevalOp;
			else
				if(eval1==eval&&bestevalOp<evalOp)
				bestmove=m,evalOp=bestevalOp;
		}
	}

	//Correct version of 3-tier evaluation with alpha-beta?!
	void Eval5(int&bestmove,EVALOK&eval,EVALOK alpha,EVALOK beta,int depth,int player,array<int>^squares,int k,int dummyOp=-1,bool extraCode=false)
	{
		eval=bestmove=0;
		if(depth<1||k>=squares->Length)return;
#if 1//Removing this call to EvalX speeds up code 3 times but gives much worse results! Why?
		if(depth==1){
			EVAL eval1=EvalX(bestmove,squares[k],extraCode);
			eval.score[player&1]=eval1.score,eval.deleted=eval1.deleted;
			return;
		}
#endif
		for(int m=0;m<15;m++){
			if(4-m>=0&&IsEmpty(4-m,0))continue;//this move is duplicated by move 5
			if(m-10>=0&&IsEmpty(m-10,BOARD_SIZE-1))continue;//this move is duplicated by move 9
			Board Board1=*this;
#if 1
			eval=Board1.MakeMoveX(m,squares[k],player);
#else//This is equivalent to above
			eval=0;
			EVAL eval1=Board1.MakeMoveX(m,squares[k]);
			eval.score[player&1]=eval1.score,eval.deleted=eval1.deleted;
#endif
			int bestmove1;
			EVALOK eval2;
			if(depth==1)goto COMPARE;
			if(k+1>=squares->Length)goto COMPARE;
			if(dummyOp>-1){
				eval+=Board1.MakeMoveX(dummyOp,squares[k+1],player+1);
				Board1.Eval5(bestmove1,eval2,alpha,beta,depth-1,player,squares,k+2,dummyOp);
				eval+=eval2;
				goto COMPARE;
			}
			Board1.Eval5(bestmove1,eval2,alpha,beta,depth-1,player+1,squares,k+1,dummyOp);
			eval+=eval2;
COMPARE:
			if((player&1)==0&&eval>alpha)
				alpha=eval,bestmove=m;
			else
			if((player&1)==1&&eval<beta)
				beta=eval,bestmove=m;
#ifdef ALPHA_BETA
			if(alpha>=beta)break;
#endif
		}
		if((player&1)==0)eval=alpha;
		else eval=beta;
	}

/**********************************************************************************************************
	This seems to be theoretically true version. It takes into account additivity of evaluation, 
	introducing an extra parameter eva0==evaluation "above" the current node.
	One has: minimax without cutoff == alpha-beta evaluation, which means that this version is almost sure bug free
	Alpha-beta here is 30 times quicker than minimax without cutoffs!
**********************************************************************************************************/
	void Eval5X(int&bestmove,EVALOK&eval,EVALOK eval0,EVALOK alpha,EVALOK beta,int depth,int player,array<int>^squares,int k,int dummyOp=-1,bool extraCode=false)
	{
		eval=bestmove=0;
		if(depth<1||k>=squares->Length)return;
#if OBSOLETE//Removing this call to obsolete EvalX() speeds up code 2 times and gives slightly better results!
		if(depth==1){
			EVAL eval1=EvalX(bestmove,squares[k],extraCode);
			eval.score[player&1]=eval1.score,eval.deleted=eval1.deleted;
			return;
		}
#endif
		for(int m=0;m<15;m++){
			if(4-m>=0&&IsEmpty(4-m,0))continue;//this move is duplicated by move 5
			if(m-10>=0&&IsEmpty(m-10,BOARD_SIZE-1))continue;//this move is duplicated by move 9
			Board Board1=*this;
			eval=Board1.MakeMoveX(m,squares[k],player);
			eval+=eval0;
			int bestmove1;
			EVALOK eval2;
			if(depth==1)goto COMPARE;
			if(k+1>=squares->Length)goto COMPARE;
			if(dummyOp>-1){
				eval+=Board1.MakeMoveX(dummyOp,squares[k+1],player+1);
				Board1.Eval5X(bestmove1,eval2,eval,alpha,beta,depth-1,player,squares,k+2,dummyOp);
				eval+=eval2;
				goto COMPARE;
			}
			Board1.Eval5X(bestmove1,eval2,eval,alpha,beta,depth-1,player+1,squares,k+1,dummyOp);
			eval+=eval2;
COMPARE:
			if((player&1)==0&&eval>alpha)
				alpha=eval,bestmove=m;
			else
			if((player&1)==1&&eval<beta)
				beta=eval,bestmove=m;
#ifdef ALPHA_BETA//OK: minimax == alpha-beta
			if(alpha>=beta)break;
#endif
		}
		if((player&1)==0)eval=alpha;
		else eval=beta;
		eval-=eval0;//one is to return "global" eval reduced by eval0!
	}

	int Eval(Engine eng,array<int>^squares,int k,int dummyOp=-1){
		int bestmove;
		EVAL eval,evalOp;
		EVALOK evalOK;
		switch(eng.strat){
			case DUMMY:
				return 5;
			case RANDOM:
				return rand()%15;
			case EVAL2:
				Eval(bestmove,eval.score,evalOp.score,eng.depth,squares,k,dummyOp);
				return bestmove;
			case EVAL2X:
				Eval(bestmove,eval.score,evalOp.score,eng.depth,squares,k,dummyOp,true);
				return bestmove;
			case EVAL3:
				EvalX(bestmove,eval,evalOp,eng.depth,squares,k,dummyOp);
				return bestmove;
			case EVAL3X:
				EvalX(bestmove,eval,evalOp,eng.depth,squares,k,dummyOp,true);
				return bestmove;
			case EVAL2XX:
				Eval2XX(bestmove,eval.score,evalOp.score,eng.depth,squares,k,dummyOp,true);
				return bestmove;
			case EVAL3XX:
				Eval3XX(bestmove,eval,evalOp,eng.depth,squares,k,dummyOp,true);
				return bestmove;
			case EVAL4X:
				Eval4X(bestmove,eval,evalOp,eng.depth,squares,k,dummyOp,true);
				return bestmove;
			case EVAL4XX:
				Eval4XX(bestmove,eval,evalOp,eng.depth,squares,k,dummyOp,true);
				return bestmove;
			case EVAL5:
				Eval5(bestmove,evalOK,INFMINUS,INFPLUS,eng.depth,0,squares,k,dummyOp,true);
				return bestmove;
			case EVAL5X:
				Eval5X(bestmove,evalOK,0,INFMINUS,INFPLUS,eng.depth,0,squares,k,dummyOp,true);
				return bestmove;
			default:
				return -1;
		}
	}

	bool operator==(const Board&board)const{
		for(int c=0;c<BOARD_SIZE;c++){
			if(board.Height.get(c)!=Height.get(c))return false;
			for(int r=0;r<BOARD_SIZE;r++){
			if(IsEmpty(r,c))continue;
			if(board.pos[r][c]!=pos[r][c])return false;
			}
		}
		return true;
	}

	bool operator!=(const Board&board)const{
		return !(board==*this);
	}

};

struct MoveStats{
	int move,pict;
	int score;
	double time;
};

extern MoveStats History[MAX_MOVES+1];
extern int Pictures[MAX_MOVES];

#if 0//Obsolete
//Strategies: MAX -- maximize own score, DIFF -- maximize difference of scores MyScore-OpScore;
//If the result of a game were evaluated as in chess (1, 0.5, 1) then the most adequate strategy
//would be DIFF, but here MAX seems to be better, if the number of participients id big enough,
//and most of them try to use DIFF strategy. Paradox: it is possible to loose ALL of the games
//using MAX strategy, but to win the tournement nevertheless!
//In fact, the best strategy seems to be adaptive one: use MAX if the opponent uses DIFF and vice versa.
//But it is not so easy to determine opponent's strategy,except the case, where it is DUMMY.
enum STRAT{MAX,DIFF,DUMMY,RANDOM,UNKNOWN};
extern int depth;//current depth of search
extern int EvalMax[15][MAX_DEPTH],EvalDiff[15][MAX_DEPTH];//Eval...[move][depth]
//Comparing moves for MAX strategy:
int compareMax(const void *arg1, const void *arg2 );
//Comparing moves for DIFF strategy:
int compareDiff(const void *arg1, const void *arg2 );
#endif

struct Game{
	Board Board0,currBoard;
	int NPict,NMoves,K;//,*Pictures;
	int N;//number of moves already played 
	Board BStack[MAX_DEPTH];
	int Score[2];//Current global score
	//STRAT strat[2];//current strategies: MAX for me and UNKNOWN (or DUMMY) for opponent
	//bool DummyOpponent,isFirst;//If one and the same move repeats 5 or more times, suppose that the opponent is in zeitnot
	int odd;//My moves: odd or even?
	int NDummy;//Number of last repeating opponents moves; if it is big enough, one may assume
	//that the opponent is in zeitnot and uses DUMMY strategy. If so, one can effectively double the search depth.
	double timePerMove,totalTime[2];
	MoveStats*History;
	int MaxDepth;
	
	bool IsDummyOp(){return NDummy>MIN_DUMMY;}
	
	//Game(){}

	//Random game:
	Game(int P,int n,int seed=0,bool odd=0,int K=0):Board0(Board(P,seed)),NPict(P),
		NMoves(n),K(K),odd(odd),timePerMove(0.2),N(odd),MaxDepth(MAX_DEPTH),
		//Pictures(::Pictures),
		History(::History+1)
	{
		currBoard=Board0;
		if(K==0)this->K=NMoves;
		if(K<MaxDepth)MaxDepth=K;
#ifdef COMPILER_BUG
		//The next 2 loops are not executed!!!
		for(int i=0;i<K;i++)Pictures[i+N]=rand()%P+1;
		for(int i=0;i<K;i++)
			//if(Pictures[i]==0)
				printf("Pictures[%d]=%d\n",i,Pictures[i]);
#else
		//the same inline code as above, but now it is compiled OK
		Init();
#endif
		//strat[0]=MAX,strat[1]=UNKNOWN;
		Score[0]=Score[1]=0;
		History[-1].move=-1;
	}

	//So one is to inicialize Pictures[] outside of the constructor:
	void Init(){
			for(int i=0;i<K;i++)Pictures[i+N]=rand()%NPict+1;
#if 0
		for(int i=0;i<K;i++)
			//if(Pictures[i]==0)
				printf("Pictures[%d]=%d\n",i,Pictures[i]);//Here OK
#endif
	}
	
	Game(array<array<int>^>^board,array<int>^squares,int n,bool isFirst,double timePerMove):
		Board0(board),NMoves(n),odd(isFirst?0:1),timePerMove(timePerMove),K(squares->Length),N(isFirst?0:1),
		//Pictures(::Pictures),
		History(::History+1)
	{
		currBoard=Board0;
		for(int i=0;i<K;i++)Pictures[i+N]=squares[i];
		//Creators of SlideWebService interface forgot to provide
		//info about the number of different pictures. So one is to calculate a lower bound for this
		NPict=3;
		for(int r=0;r<BOARD_SIZE;r++)
		for(int c=0;c<BOARD_SIZE;c++)
			if(board[r][c]>NPict)NPict=board[r][c];
		for(int i=0;i<K;i++)
			if(squares[i]>NPict)NPict=squares[i];
		//strat[0]=MAX,strat[1]=UNKNOWN;
		Score[0]=Score[1]=0;
		History[-1].move=-1;
#ifdef TEST
		printf("Initial position:\n\n");
		Board0.Print();
		printf("NMoves=%d K=%d NPict=%d isFirst=%s\n",NMoves,K,NPict,odd==0?"true":"false");
#endif
	}

	void RunGreedy(){
#ifdef PRINT_POSITIONS
		printf("Greedy search:\n\n");
#endif
		for(int n=0;n<NMoves;n++){
			EVAL eval;
			int bestmove=0;
			eval=currBoard.EvalX(bestmove,Pictures[n]);
			currBoard.MakeMoveX(bestmove,Pictures[n]);
			Score[n%2]+=eval.score;
#ifdef PRINT_POSITIONS
			printf("%d. Pict %d, BestMove %d. Eval %d, Score %d. Position:\n",n,Pictures[n],bestmove,eval.score,Score[n%2]);
			currBoard.Print();
#endif
		}
		printf("Score: %d : %d\n",Score[0],Score[1]);
	}

	void Run(array<int>^squares,int depth,int depthOp,bool extraCode=false)
	{
#ifdef TEST
			printf("Search at depth %d:%d\n\n",depth,depthOp);
#endif
		for(int n=0;n<NMoves;n++){
			EVAL eval=0,evalOp=0;
			int bestmove=0;
			currBoard.EvalX(bestmove,eval,evalOp,(n&1)==0?depth:depthOp,squares,n,-1,extraCode);
			int deltascore=currBoard.MakeMoveX(bestmove,squares[n]);
			Score[n%2]+=deltascore;
#ifdef PRINT_POSITIONS
			printf("%d. Pict %d, BestMove %d. Eval %d, deltaScore %d, Score %d. Position:\n",n,squares[n],bestmove,eval.score,deltascore,Score[n%2]);
			currBoard.Print();
#endif
		}
		printf("Score: %d : %d\n",Score[0],Score[1]);
	}

	void Run(array<int>^squares,Engine eng,Engine engOp,double*times=0)
	{
#ifdef TEST
		//printf("Engines %s:%s at depth %d:%d\n\n",EngineNames[eng.strat],EngineNames[engOp.strat],eng.depth,engOp.depth);
#endif
		int dummyOp=engOp.strat==0?5:-1,dummyOpOp=eng.strat==0?5:-1;
		clock_t start0=clock(),start,totals[2];
		totals[0]=totals[1]=0;
		for(int n=0;n<NMoves;n++){
			start=clock();
			int bestmove=currBoard.Eval((n&1)==0?eng:engOp,squares,n,(n&1)==0?dummyOp:dummyOpOp);
			totals[n%2]+=clock()-start;
			int deltascore=currBoard.MakeMoveX(bestmove,squares[n]);
			Score[n%2]+=deltascore;
#ifdef PRINT_POSITIONS
			printf("%d. Pict %d, BestMove %d. deltaScore %d, Score %d. Position:\n",n,squares[n],bestmove,deltascore,Score[n%2]);
			currBoard.Print();
#endif
		}
		printf("Score: %d : %d\n",Score[0],Score[1]);
		printf("Times used: %f:%f sec\n",(double)totals[0]/CLOCKS_PER_SEC,(double)totals[1]/CLOCKS_PER_SEC);
		if(times!=0)
			times[0]=(double)totals[0]/CLOCKS_PER_SEC,times[1]=(double)totals[1]/CLOCKS_PER_SEC;
	}

};
